@@ -9,18 +9,19 @@ specified in the meta-information lines --  specifically the ##INFO and
 against the reserved types mentioned in the spec.  Failing that, it will just
 return strings.
 
-There is currently one piece of interface: ``VCFReader``.  It takes a file-like
+There is currently one piece of interface: ``Reader``.  It takes a file-like
 object and acts as a reader::
 
     >>> import vcf
-    >>> vcf_reader = vcf.VCFReader(open('example.vcf', 'rb'))
+    >>> vcf_reader = vcf.Reader(open('test/example.vcf', 'rb'))
     >>> for record in vcf_reader:
     ...     print record
-    Record(CHROM='20', POS=14370, ID='rs6054257', REF='G', ALT=['A'], QUAL=29,
-    FILTER='PASS', INFO={'H2': True, 'NS': 3, 'DB': True, 'DP': 14, 'AF': [0.5]
-    }, FORMAT='GT:GQ:DP:HQ', samples=[{'GT': '0', 'HQ': [58, 50], 'DP': 3, 'GQ'
-    : 49, 'name': 'NA00001'}, {'GT': '0', 'HQ': [65, 3], 'DP': 5, 'GQ': 3, 'nam
-    e' : 'NA00002'}, {'GT': '0', 'DP': 3, 'GQ': 41, 'name': 'NA00003'}])
+    Record(CHROM=20, POS=14370, REF=G, ALT=['A'])
+    Record(CHROM=20, POS=17330, REF=T, ALT=['A'])
+    Record(CHROM=20, POS=1110696, REF=A, ALT=['G', 'T'])
+    Record(CHROM=20, POS=1230237, REF=T, ALT=['.'])
+    Record(CHROM=20, POS=1234567, REF=GTCT, ALT=['G', 'GTACT'])
+
 
 This produces a great deal of information, but it is conveniently accessed.
 The attributes of a Record are the 8 fixed fields from the VCF spec plus two
@@ -35,63 +36,89 @@ more.  That is:
     * ``Record.FILTER``
     * ``Record.INFO``
 
-plus two more attributes to handle genotype information:
+plus three more attributes to handle genotype information:
 
     * ``Record.FORMAT``
     * ``Record.samples``
+    * ``Record.genotype``
 
-``samples``, not being the title of any column, is left lowercase.  The format
+``samples`` and ``genotypes``, not being the title of any column, is left lowercase.  The format
 of the fixed fields is from the spec.  Comma-separated lists in the VCF are
 converted to lists.  In particular, one-entry VCF lists are converted to
 one-entry Python lists (see, e.g., ``Record.ALT``).  Semicolon-delimited lists
 of key=value pairs are converted to Python dictionaries, with flags being given
 a ``True`` value. Integers and floats are handled exactly as you'd expect::
 
+    >>> vcf_reader = vcf.Reader(open('test/example.vcf', 'rb'))
     >>> record = vcf_reader.next()
     >>> print record.POS
-    17330
+    14370
     >>> print record.ALT
     ['A']
     >>> print record.INFO['AF']
-    [0.017]
+    [0.5]
 
 ``record.FORMAT`` will be a string specifying the format of the genotype
 fields.  In case the FORMAT column does not exist, ``record.FORMAT`` is
 ``None``.  Finally, ``record.samples`` is a list of dictionaries containing the
-parsed sample column::
+parsed sample column and ``record.genotype`` is a way of looking up genotypes
+by sample name::
 
     >>> record = vcf_reader.next()
     >>> for sample in record.samples:
     ...     print sample['GT']
-    '1|2'
-    '2|1'
-    '2/2'
+    0|0
+    0|1
+    0/0
+    >>> print record.genotype('NA00001')['GT']
+    0|0
 
 Metadata regarding the VCF file itself can be investigated through the
 following attributes:
 
-    * ``VCFReader.metadata``
-    * ``VCFReader.infos``
-    * ``VCFReader.filters``
-    * ``VCFReader.formats``
-    * ``VCFReader.samples``
+    * ``Reader.metadata``
+    * ``Reader.infos``
+    * ``Reader.filters``
+    * ``Reader.formats``
+    * ``Reader.samples``
 
 For example::
 
     >>> vcf_reader.metadata['fileDate']
-    20090805
+    '20090805'
     >>> vcf_reader.samples
     ['NA00001', 'NA00002', 'NA00003']
     >>> vcf_reader.filters
-    {'q10': Filter(id='q10', desc='Quality below 10'),
-    's50': Filter(id='s50', desc='Less than 50% of samples have data')}
+    {'q10': Filter(id='q10', desc='Quality below 10'), 's50': Filter(id='s50', desc='Less than 50% of samples have data')}
     >>> vcf_reader.infos['AA'].desc
-    Ancestral Allele
+    'Ancestral Allele'
+
+Random access is supported for files with tabix indexes.  Simply call fetch for the
+region you are interested in::
+
+    >>> vcf_reader = vcf.Reader(filename='test/tb.vcf.gz')
+    >>> for record in vcf_reader.fetch('20', 1110696-1, 1230237):
+    ...     print record
+    Record(CHROM=20, POS=1110696, REF=A, ALT=['G', 'T'])
+    Record(CHROM=20, POS=1230237, REF=T, ALT=['.'])
+
+
+An extensible script is available to filter vcf files in vcf_filter.py.  VCF filters
+declared by other packages will be available for use in this script.  Please
+see FILTERS.md for full description.
 
 '''
 import collections
 import re
 import csv
+import gzip
+import itertools
+
+
+try:
+    import pysam
+except ImportError:
+    pysam = None
 
 
 # Metadata parsers/constants
@@ -120,7 +147,7 @@ class _vcf_metadata_parser(object):
         self.aggro = aggressive
         self.info_pattern = re.compile(r'''\#\#INFO=<
             ID=(?P<id>[^,]+),
-            Number=(?P<number>\d+|\.|[AG]),
+            Number=(?P<number>-?\d+|\.|[AG]),
             Type=(?P<type>Integer|Float|Flag|Character|String),
             Description="(?P<desc>[^"]*)"
             >''', re.VERBOSE)
@@ -130,7 +157,7 @@ class _vcf_metadata_parser(object):
             >''', re.VERBOSE)
         self.format_pattern = re.compile(r'''\#\#FORMAT=<
             ID=(?P<id>.+),
-            Number=(?P<number>\d+|\.|[AG]),
+            Number=(?P<number>-?\d+|\.|[AG]),
             Type=(?P<type>.+),
             Description="(?P<desc>.*)"
             >''', re.VERBOSE)
@@ -145,6 +172,8 @@ class _vcf_metadata_parser(object):
 
         try:
             num = int(match.group('number'))
+            if num < 0:
+                num = None if self.aggro else '.'
         except ValueError:
             num = None if self.aggro else '.'
 
@@ -175,6 +204,8 @@ class _vcf_metadata_parser(object):
 
         try:
             num = int(match.group('number'))
+            if num < 0:
+                num = None if self.aggro else '.'
         except ValueError:
             num = None if self.aggro else '.'
 
@@ -188,116 +219,89 @@ class _vcf_metadata_parser(object):
         return match.group('key'), match.group('val')
 
 
-# Reader class
-class _meta_info(object):
-    '''Decorator for a property stored in the header info.'''
-    def __init__(self, func):
-        self.func = func
-
-    def __call__(self, fself):
-        if getattr(fself, "_%s" % self.func.__name__) is None:
-            fself._parse_metainfo()
-
-        return self.func(fself)
-
-    def __repr__(self):
-        '''Return the function's docstring.'''
-        return self.func.__doc__
-
-    def __doc__(self):
-        '''Return the function's docstring.'''
-        return self.func.__doc__
-
 
 class _Record(object):
-    def __init__(self, CHROM, POS, ID, REF, ALT, QUAL, FILTER, INFO, FORMAT, samples):
+    def __init__(self, CHROM, POS, ID, REF, ALT, QUAL, FILTER, INFO, FORMAT, samples, sample_indexes):
         self.CHROM = CHROM
         self.POS = POS
         self.ID = ID
         self.REF = REF
-        self.ALT = ALT 
+        self.ALT = ALT
         self.QUAL = QUAL
         self.FILTER = FILTER
         self.INFO = INFO
         self.FORMAT = FORMAT
         self.samples = samples
+        self._sample_indexes = sample_indexes
 
     def __str__(self):
-        return "%(CHROM)s\t%(POS)s\t%(REF)s\t%(ALT)s" % self.__dict__
+        return "Record(CHROM=%(CHROM)s, POS=%(POS)s, REF=%(REF)s, ALT=%(ALT)s)" % self.__dict__
 
-    def sample_lookup(self):
-        return dict(((x['name'], x) for x in self.samples))
-        
     def add_format(self, fmt):
         self.FORMAT = self.FORMAT + ':' + fmt
 
     def add_filter(self, flt):
-        if self.FILTER == '.':
+        if self.FILTER == '.' or self.FILTER == 'PASS':
             self.FILTER = ''
-        else: 
+        else:
             self.FILTER = self.FILTER + ';'
         self.FILTER = self.FILTER + flt
-    
+
     def add_info(self, info, value=True):
         self.INFO[info] = value
 
-        
-class VCFReader(object):
+
+    def genotype(self, name):
+        return self.samples[self._sample_indexes[name]]
+
+
+class Reader(object):
     '''Read and parse a VCF v 4.0 file'''
-    def __init__(self, fsock, aggressive=False):
+    def __init__(self, fsock=None, filename=None, aggressive=False, compressed=False):
         super(VCFReader, self).__init__()
+
+        if not (fsock or filename):
+            raise Exception('You must provide at least fsock or filename')
+
+        if filename:
+            self.filename = filename
+            if fsock is None:
+                self.reader = file(filename)
+
+        if fsock:
+            self.reader = fsock
+            if filename is None:
+                if hasattr(fsock, 'name'):
+                    filename = fsock.name
+            self.filename = filename
+
+        if compressed or (filename and filename.endswith('.gz')):
+            self.reader = gzip.GzipFile(fileobj=self.reader)
+
         self.aggro = aggressive
-        self._metadata = None
-        self._infos = None
-        self._filters = None
-        self._formats = None
-        self._samples = None
-        self.reader = fsock
+        self.metadata = None
+        self.infos = None
+        self.filters = None
+        self.formats = None
+        self.samples = None
+        self._sample_indexes = None
         self._header_lines = []
+        self._tabix = None
         if aggressive:
             self._mapper = self._none_map
         else:
             self._mapper = self._pass_map
+        self._parse_metainfo()
 
     def __iter__(self):
         return self
 
-    @property
-    @_meta_info
-    def metadata(self):
-        '''Return the information from lines starting "##"'''
-        return self._metadata
-
-    @property
-    @_meta_info
-    def infos(self):
-        '''Return the information from lines starting "##INFO"'''
-        return self._infos
-
-    @property
-    @_meta_info
-    def filters(self):
-        '''Return the information from lines starting "##FILTER"'''
-        return self._filters
-
-    @property
-    @_meta_info
-    def formats(self):
-        '''Return the information from lines starting "##FORMAT"'''
-        return self._formats
-
-    @property
-    @_meta_info
-    def samples(self):
-        '''Return the names of the genotype fields.'''
-        return self._samples
-
     def _parse_metainfo(self):
         '''Parse the information stored in the metainfo of the VCF.
 
         The end user shouldn't have to use this.  She can access the metainfo
         directly with ``self.metadata``.'''
-        for attr in ('_metadata', '_infos', '_filters', '_formats'):
+        for attr in ('metadata', 'infos', 'filters', 'formats'):
             setattr(self, attr, {})
 
         parser = _vcf_metadata_parser()
@@ -306,27 +310,28 @@ class VCFReader(object):
         while line.startswith('##'):
             self._header_lines.append(line)
             line = line.strip()
-            
+
             if line.startswith('##INFO'):
                 key, val = parser.read_info(line)
-                self._infos[key] = val
+                self.infos[key] = val
 
             elif line.startswith('##FILTER'):
                 key, val = parser.read_filter(line)
-                self._filters[key] = val
+                self.filters[key] = val
 
             elif line.startswith('##FORMAT'):
                 key, val = parser.read_format(line)
-                self._formats[key] = val
+                self.formats[key] = val
 
             else:
                 key, val = parser.read_meta(line.strip())
-                self._metadata[key] = val
+                self.metadata[key] = val
 
             line = self.reader.next()
 
         fields = line.split()
-        self._samples = fields[9:]
+        self.samples = fields[9:]
+        self._sample_indexes = dict([(x,i) for (i,x) in enumerate(self.samples)])
 
     def _none_map(self, func, iterable, bad='.'):
         '''``map``, but make bad values None.'''
@@ -383,14 +388,19 @@ class VCFReader(object):
     def _parse_samples(self, samples, samp_fmt):
         '''Parse a sample entry according to the format specified in the FORMAT
         column.'''
-        samp_data = []
+        samp_data = []# OrderedDict()
         samp_fmt = samp_fmt.split(':')
-        for sample in samples:
-            sampdict = dict(zip(samp_fmt, sample.split(':')))
+
+        # cache ref to speed loop
+        formats = self.formats
+        mapper = self._mapper
+
+        for name, sample in itertools.izip(self.samples, samples):
+            sampdict = dict(itertools.izip(samp_fmt, sample.split(':')))
             for fmt in sampdict:
                 vals = sampdict[fmt].split(',')
                 try:
-                    entry_type = self.formats[fmt].type
+                    entry_type = formats[fmt].type
                 except KeyError:
                     try:
                         entry_type = RESERVED_FORMAT[fmt]
@@ -398,27 +408,19 @@ class VCFReader(object):
                         entry_type = 'String'
 
                 if entry_type == 'Integer':
-                    sampdict[fmt] = self._mapper(int, vals)
+                    sampdict[fmt] = mapper(int, vals)
                 elif entry_type == 'Float' or entry_type == 'Numeric':
-                    sampdict[fmt] = self._mapper(float, vals)
+                    sampdict[fmt] = mapper(float, vals)
                 elif sampdict[fmt] == './.' and self.aggro:
                     sampdict[fmt] = None
 
+            sampdict['name'] = name
             samp_data.append(sampdict)
 
-        for name, data in zip(self.samples, samp_data):
-            data['name'] = name
-            if data['GT'] != '.':
-                data['_GT'] = int(data['GT'][0]), int(data['GT'][-1])
-            else: 
-                data['_GT'] = None
-                
         return samp_data
 
     def next(self):
         '''Return the next record in the file.'''
-        if self._samples is None:
-            self._parse_metainfo()
         row = self.reader.next().split()
         chrom = row[0]
         pos = int(row[1])
@@ -430,7 +432,11 @@ class VCFReader(object):
 
         ref = row[3]
         alt = self._mapper(str, row[4].split(','))
-        qual = float(row[5]) if '.' in row[5] else int(row[5])
+
+        if row[5] == '.':
+            qual = '.'
+        else:
+            qual = float(row[5]) if '.' in row[5] else int(row[5])
         filt = row[6].split(';') if ';' in row[6] else row[6]
         if filt == 'PASS' and self.aggro:
             filt = None
@@ -445,89 +451,115 @@ class VCFReader(object):
             samples = self._parse_samples(row[9:], fmt)
 
         record = _Record(chrom, pos, ID, ref, alt, qual, filt, info, fmt,
-                         samples)
+                         samples, self._sample_indexes)
         return record
 
+    def fetch(self, chrom, start, end):
+        if not pysam:
+            raise Exception('pysam not available, try "pip install pysam"?')
+
+        if not self.filename:
+            raise Exception('Please provide a filename (or a "normal" fsock)')
+
+        if not self._tabix:
+            self._tabix = pysam.Tabixfile(self.filename)
+
+        self.reader = self._tabix.fetch(chrom, start, end)
+        return self
+
+class Writer(object):
 
-class VCFWriter(object):
-    
     fixed_fields = "#CHROM POS ID REF ALT QUAL FILTER INFO FORMAT".split()
-    
+
     def __init__(self, stream, template):
         self.writer = csv.writer(stream, delimiter="\t")
         self.template = template
-        # TODO: shouldnt have to poke the parser to get the meta
-        if template._samples is None: 
-            template._parse_metainfo()
-        for line in template._header_lines: 
-            stream.write(line)
+
+        for line in template.metadata.items():
+            stream.write('##%s=%s\n' % line)
+        for line in template.infos.values():
+            stream.write('##INFO=<ID=%s,Number=%s,Type=%s,Description="%s">\n' % line)
+        for line in template.formats.values():
+            stream.write('##FORMAT=<ID=%s,Number=%s,Type=%s,Description="%s">\n' % line)
+
+        for line in template.filters.values():
+            stream.write('##FILTER=<ID=%s,Description="%s">\n' % line)
+
+        self.info_pattern = re.compile(r'''\#\#INFO=<
+            ID=(?P<id>[^,]+),
+            Number=(?P<number>-?\d+|\.|[AG]),
+            Type=(?P<type>Integer|Float|Flag|Character|String),
+            Description="(?P<desc>[^"]*)"
+            >''', re.VERBOSE)
+        self.filter_pattern = re.compile(r'''\#\#FILTER=<
+            ID=(?P<id>[^,]+),
+            Description="(?P<desc>[^"]*)"
+            >''', re.VERBOSE)
+        self.format_pattern = re.compile(r'''\#\#FORMAT=<
+            ID=(?P<id>.+),
+            Number=(?P<number>-?\d+|\.|[AG]),
+            Type=(?P<type>.+),
+            Description="(?P<desc>.*)"
+            >''', re.VERBOSE)
+        self.meta_pattern = re.compile(r'''##(?P<key>.+)=(?P<val>.+)''')
+
         self.write_header()
-        
+
     def write_header(self):
         # TODO: write INFO, etc
         self.writer.writerow(self.fixed_fields + self.template.samples)
 
     def write_record(self, record):
-        ffs = [record.CHROM, record.POS, record.ID, record.REF, self._format_alt(record.ALT), 
+        ffs = [record.CHROM, record.POS, record.ID, record.REF, self._format_alt(record.ALT),
             record.QUAL, record.FILTER, self._format_info(record.INFO), record.FORMAT]
-        
-        samples = [self._format_sample(record.FORMAT, sample) 
+
+        samples = [self._format_sample(record.FORMAT, sample)
             for sample in record.samples]
         self.writer.writerow(ffs + samples)
 
     def _format_alt(self, alt):
         return ','.join(alt)
-    
+
     def _format_info(self, info):
         return ';'.join(["%s=%s" % (x, self._stringify(y)) for x, y in info.items()])
 
     def _format_sample(self, fmt, sample):
-        if sample["GT"] == ".":
-            return "."
+        if sample["GT"] == "./.":
+            return "./."
         return ':'.join((str(self._stringify(sample[f])) for f in fmt.split(':')))
 
     def _stringify(self, x):
-        if type(x) == type([]): 
+        if type(x) == type([]):
             return ','.join(map(str, x))
         return str(x)
 
-def main():
-    '''Parse the example VCF file from the specification and print every
-    record.'''
-    import contextlib
-    import StringIO
-    import textwrap
-    buff = '''\
-        ##fileformat=VCFv4.0
-        ##fileDate=20090805
-        ##source=myImputationProgramV3.1
-        ##reference=1000GenomesPilot-NCBI36
-        ##phasing=partial
-        ##INFO=<ID=NS,Number=1,Type=Integer,Description="Number of Samples With Data">
-        ##INFO=<ID=DP,Number=1,Type=Integer,Description="Total Depth">
-        ##INFO=<ID=AF,Number=.,Type=Float,Description="Allele Frequency">
-        ##INFO=<ID=AA,Number=1,Type=String,Description="Ancestral Allele">
-        ##INFO=<ID=DB,Number=0,Type=Flag,Description="dbSNP membership, build 129">
-        ##INFO=<ID=H2,Number=0,Type=Flag,Description="HapMap2 membership">
-        ##INFO=<ID=AC,Number=A,Type=Integer,Description="Total number of alternate alleles in called genotypes">
-        ##FILTER=<ID=q10,Description="Quality below 10">
-        ##FILTER=<ID=s50,Description="Less than 50% of samples have data">
-        ##FORMAT=<ID=GT,Number=1,Type=String,Description="Genotype">
-        ##FORMAT=<ID=GQ,Number=1,Type=Integer,Description="Genotype Quality">
-        ##FORMAT=<ID=DP,Number=1,Type=Integer,Description="Read Depth">
-        ##FORMAT=<ID=HQ,Number=2,Type=Integer,Description="Haplotype Quality">
-        #CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\tNA00001\tNA00002\tNA00003
-        20\t14370\trs6054257\tG\tA\t29\tPASS\tNS=3;DP=14;AF=0.5;DB;H2\tGT:GQ:DP:HQ\t0|0:48:1:51,51\t1|0:48:8:51,51\t1/1:43:5:.,.
-        20\t17330\t.\tT\tA\t3\tq10\tNS=3;DP=11;AF=0.017\tGT:GQ:DP:HQ\t0|0:49:3:58,50\t0|1:3:5:65,3\t0/0:41:3
-        20\t1110696\trs6040355\tA\tG,T\t67\tPASS\tNS=2;DP=10;AF=0.333,0.667;AA=T;DB\tGT:GQ:DP:HQ\t1|2:21:6:23,27\t2|1:2:0:18,2\t2/2:35:4
-        20\t1230237\t.\tT\t.\t47\tPASS\tNS=3;DP=13;AA=T\tGT:GQ:DP:HQ\t0|0:54:7:56,60\t0|0:48:4:51,51\t0/0:61:2
-        20\t1234567\tmicrosat1\tGTCT\tG,GTACT\t50\tPASS\tNS=3;DP=9;AA=G\tGT:GQ:DP\t./.:35:4\t0/2:17:2\t1/1:40:3
-        '''
-    with contextlib.closing(StringIO.StringIO(textwrap.dedent(buff))) as sock:
-        vcf_file = VCFReader(sock, aggressive=True)
-        for record in vcf_file:
-            print record
 
 
-if __name__ == '__main__':
-    main()
+class Filter(object):
+    name = 'filter'
+    description = 'VCF filter base class'
+    short_name = 'f'
+
+    @classmethod
+    def customize_parser(self):
+        pass
+
+    def __init__(self, args):
+        self.threshold = 0
+
+    def __call__(self):
+        raise NotImplementedError('Filters must implement this method')
+
+    def filter_name(self):
+        return '%s%s' % (self.short_name, self.threshold)
+
+
+def __update_readme():
+    import sys
+    file('README.rst', 'w').write(sys.modules[__name__].__doc__)
+
+
+# backwards compatibility
+VCFReader = Reader
+VCFWriter = Writer
+
